/*
** Convert
**
** Tiny program to convert 256 color BMPs in some binary formats
** for being included in C
**
** These include:
**  RCKye intensity maps
**  PPC - PCX raw data
**  Palette data
**
** Outputs to stdout
**
** RCKye intensity map format:
**
**  Header: 1 byte, specifiing no. of images -1 (1-256)
**  Image data:
**   1 byte 'copy from' data
**    bit 0: 1 if copied, then next byte is the image index to copy
**    bit 1: Mirror X flag
**    bit 2: Mirror Y flag
**    bit 3: Mirror on the X=Y line flag
**           (Always used on the image data as a "postprocessing")
**   1 byte packed xy size (or image index if copied)
**    Top 4 bits: x-1, bottom 4 bits: y-1 (so both from 1-16)
**   Raw image data:
**    Top 4 bits: intensity
**    Bottom 4 bits: repeat count -1 (1-16)
**    (Processed until it reaches the image size calculated from xy)
**
**  Conversion to this works by reading the first line of the bmp seeking
**  for black dots (0x000000 palette entry) which show where new images
**  should begin. The actual decoding starts with the second line of the
**  image. The height of the images will be one less than the bmp's height
**  or 16 if it is taller.
**
** PPC - PCX:
**  First 2 bytes: y size (0-65535), MSB first
**  After PPC - PCX data like how described in Panthera
**
** Palette data:
**  First byte: no. of colors -1 (1-256)
**  Standard RGB 24bit palette, R, G, B order
**
**
** Calling:
**  ppciconv file1 file2 ...
**   creates palette data & converts to PPC - PCX,
**  ppciconv -p file1 file2 ...
**   creates palette data from all
**  ppciconv -i file1 file2 ...
**   creates intensity map
**  ppciconv -c file1 file2 ...
**   creates PPC PCX data
**
** Ignores any image which is not suitable
*/

#include <stdio.h>
#include <stdlib.h>



/*Some binary tweaks where no initial value is necessary
**(To stop annoying GCC complaints as it gets lost in the IFs)*/
#define INT_RUCH 'R'+('u'<<8)+('C'<<16)+('h'<<24)




/*Returns 6(+1) byte hex string:
** " 0xHH,"*/
char hexout_buf[7]={' ','0','x','4','3',',',0};
char* HexOut(unsigned char c)
{
 if ((c>>4)<10) hexout_buf[3]='0'+(c>>4);
 else hexout_buf[3]='A'+(c>>4)-10;
 if ((c&15)<10) hexout_buf[4]='0'+(c&15);
 else hexout_buf[4]='A'+(c&15)-10;
 return hexout_buf;
}




int main (int argn, char* argv[])
{
 unsigned char intmap[256]={ /*Just for a secret :lol:*/
 0x0D,0x0A,0x0D,0x0A,  /*4*/
 ' ', ' ', ' ', ' ', ' ', '<', '\\','-', ',', 0x0D,0x0A,   /*11*/
 ' ', ' ', ' ', '.', '/', ' ', '\\','_', '_', '>', ' ', ' ', 'R', 'u', 'C', 'h', 0x0D,0x0A,  /*18*/
 ' ', ' ', '/', '\'',' ', ' ', '<', 0x0D,0x0A,   /*9*/
 ' ', '|', ' ', ',', ',', ' ', ' ', '\\',' ', 'r', 'u', 'c', 'h', '.', 'c', 'o', '.', 'n', 'r', 0x0D,0x0A,  /*21*/
 ' ', '|', '/', '\'','`', '\\',' ', '>', '|', 0x0D,0x0A,  /*11*/
 ' ', '|', '|', '\\','/', '|', '|', ' ', '`', '\\',' ', ' ', '<', '~', '\\',' ', ' ', 'F', 0x0D,0x0A,  /*20*/
 ' ', '|', '|', '`', '\\','|', '|', ' ', ' ', '`', '\\',',', ' ', '~', '\\','|', ' ', 'r', 0x0D,0x0A,  /*20*/
 ' ', '|', '|', ' ', '`', '|', '|', '\\','_', '(', ' ', ' ', '\\',' ', '/', '|', ' ', 'e', 0x0D,0x0A,  /*20*/
 ' ', '|', '|', ' ', ' ', '|', '|', '_', '\\',' ', '\\',' ', '|', '~', '/', '\'',' ', 'e', 0x0D,0x0A,  /*20*/
 '(', 'W', ')', ' ', '(', 'W', ')', '_', '(', 'W', '_', '_', '|', '~', 0x0D,0x0A,  /*16*/
 /*170*/
 0x0D,0x0A,  /*2*/
 'Y','o','u',' ','a','r','e',' ','n','o','t',' ','a','n',' ',  /*15*/
 'A','S','M',' ','g','u','r','u',',',' ','o','k','?','?','!',  /*15*/
 0x0D,0x0A,  /*2*/
 'S','t','o','p',' ','m','e','s','s','i','n','g',' ','w','i','t','h',' ','t','h','i','s','!',  /*23*/
 ' ',':','3',  /*3*/
 /*60 + 170 = 230*/
 0x0D,0x0A,0x0D,0x0A,  /*4*/
 'P','r','o','u','d',' ','t','o',' ','b','e',' ','F','u','r','r','y','!', /*18*/
 0x0D,0x0A,0x0D,0x0A};
/*      <\-,
**    ./ \__>  RuCh
**   /'  <
**  | ,,  \ ruch.co.nr
**  |/'`\ >|
**  ||\/|| `\  <~\  F
**  ||`\||  `\, ~\| r
**  || `||\_(  \ /| e
**  ||  ||_\ \ |~/' e
** (W) (W)_(W__|~
** You are not an ASM guru, ok??!
** Stop messing with this! :3
** 
** Proud to be Furry! */
 int cpal=1, cppc=1, cint=0, zero=-1, transl=-1;
 int carg=1;
 FILE *fh;
 int t1,t2,t3,t4,t5,t6,t7,t8,osize=INT_RUCH,oasize=0,intmapsc;
 unsigned char *bmp=NULL, *bmporig=NULL, *in, *intmaps=NULL, *intmapst=NULL, *intmapsx=NULL, *intmapsy=NULL, *intmapsm=NULL, *intmapscopy=NULL;
 unsigned int ioff=INT_RUCH;
 int bmpl=INT_RUCH,sx=INT_RUCH,sxa=INT_RUCH,sy=INT_RUCH,psize=INT_RUCH,black=0;

 /*No files given: output help*/
 if ((argn==1) || ((argn==2) && (argv[1][0]=='-'))){
  printf("PPC Include Converter 1.00\nRunning Cheetah, July, 2006\nConverts 256 color BMP images to some includeable hex dumps\n\nUsage:\n\n ppciconv [-p|-i[index]|-(c|C)[index]] file1.bmp [file2.bmp [...]]\n\nby default creates palette data & converts to PPC - PCX.\n-p: Only create palette data\n-i: Only create intensity map (black seperates)\n    If index is given, that color will be used as seperator\n-c: Only create PPC - PCX image data\n-C: Create palette and PPC - PCX image data (default)\n    If index is given, that color will be treated as transparent\n(for creating those weird intensity maps, see source)\n");
  return 0;
 }
 
 /*Processing parameter if any*/
 if (argv[1][0]=='-'){
  if (argv[1][2]){ /*A number might be here*/
   sscanf(argv[1]+2,"%d",&t1);
  }else t1=-1;
  switch(argv[1][1]){
   case 'i': cpal=0; cint=1; zero=t1;
             if ((intmaps=malloc(16*16*262))==NULL){ printf("Sorry, no memory for generating intensity maps\n"); return 0;}
	     intmapst=intmaps+65536; intmapsx=intmapst+512; intmapsy=intmapst+768; intmapsm=intmapst+1024; intmapscopy=intmapst+1280;
   case 'p': cppc=0; break;
   case 'c': cpal=0;
   case 'C': transl=t1;
   default:;
  }
  carg++;
 }
 
 /*Cycling through output files & generating stuff*/
 while (carg<argn){

  if (NULL==(fh=fopen(argv[carg],"rb"))) printf("/*Unable to open: %s*/\n\n",argv[carg]);
  else{
   
   /*Reading in the file*/
   t2=0;
   if (fseek(fh,0,SEEK_END)!=0) t2=1; /*To get file size*/
   else{
    bmpl=(int)ftell(fh);
    fseek(fh,0,SEEK_SET); /*Seek back to beginning*/
    if (NULL==(bmp=malloc(bmpl))) t2=1;
    else if (fread(bmp,(size_t)(bmpl),1,fh)==0){ t2=1; free(bmp); }
   }
   fclose(fh);
   bmporig=bmp;

   if (t2){ printf("/*Could not read in: %s*/\n\n",argv[carg]); carg++; continue; }
      
   /*Take a look at the file... Is it suitable?*/   
   if (bmpl<54) t2=1;
   else{

    /*14 byte file header*/
    if (((*bmp)!='B') || (bmp[1]!='M')) t2=1;
    bmp+=2;
    t1=(*bmp)+(bmp[1]<<8)+(bmp[2]<<16)+(bmp[3]<<24);
    if (bmpl<t1) t2=1;
    else{
     bmp+=8; /*Pass over a reserved dword*/
     ioff=(*bmp)+(bmp[1]<<8)+(bmp[2]<<16)+(bmp[3]<<24);
     if (bmpl<=ioff) t2=1;
     else{
      bmp+=4;
      bmpl-=14;
      ioff-=14;
     }
    }
   }
   if (t2) {printf("/*Not a valid BitMap: %s*/\n\n",argv[carg]); carg++; continue; }
   
   /*Process the info header*/
   t3=(*bmp)+(bmp[1]<<8)+(bmp[2]<<16)+(bmp[3]<<24);
   if (ioff<t3) t2=2; /*Corrupted*/
   else if (t3<40) t2=1; /*Unknown*/
   else{
    bmp+=4; 
    sx=(*bmp)+(bmp[1]<<8)+(bmp[2]<<16)+(bmp[3]<<24); bmp+=4;
    sy=(*bmp)+(bmp[1]<<8)+(bmp[2]<<16)+(bmp[3]<<24); bmp+=4;
    t1=(*bmp)+(bmp[1]<<8);
    if (t1!=1) t2=1; /*No. of planes... Must == 1*/
    else{
     bmp+=2;
     t1=(*bmp)+(bmp[1]<<8);
     if (t1!=8) t2=1; /*Not 256 color*/
     else{
      bmp+=2;
      t1=(*bmp)+(bmp[1]<<8)+(bmp[2]<<16)+(bmp[3]<<24);
      if (t1) t2=1; /*Compressed*/
      else{
       bmp+=16;
       psize=(*bmp)+(bmp[1]<<8)+(bmp[2]<<16)+(bmp[3]<<24);
       if (psize==0) psize=256;
       else if (psize>256) t2=2; /*Erratic*/
       if (t2==0){
	bmp+=t3-32;
	bmpl-=t3;
	ioff-=t3;
	if (ioff<(psize<<2)) t2=2; /*No space left for palette!*/
       }
      }
     }
    }
   }
   if (t2){
    if (t2==1) printf("/*Not a 256 color BitMap: %s*/\n\n",argv[carg]);
    else printf("/*This BitMap is corrupted: %s*/\n\n",argv[carg]);
    carg++; continue;
   }
   
   /*Read in palette & process when necessary*/
   if (cpal){ osize=1; printf("/*%s palette data begins*/\n\n%s\n",argv[carg],HexOut(psize-1)); }
   for (t1=0;t1<psize;t1++){
    intmap[t1]=(((255-bmp[2])*5)+((255-bmp[1])*7)+((255-bmp[0])*4))>>8; /*0-15 intensity*/
    if ((*((int*)bmp))==0) black=t1; /*Black color: needed for intensity maps*/
    if (cpal){ /*Save palette*/
     printf("%s",HexOut(bmp[2]));
     printf("%s",HexOut(bmp[1]));
     printf("%s",HexOut(bmp[0]));
     printf("\n"); osize+=3;
    }
    bmp+=4;
   }
   if (cpal){ printf("\n/*End of %d bytes*/\n\n\n",osize); oasize+=osize; }
   ioff-=(psize<<2);
   bmpl-=(psize<<2);
   bmp+=ioff;
   bmpl-=ioff;
    
   sxa=((sx+3) & 0xFFFFFFFC);
   if (sxa*sy>bmpl){ printf("/*The image data in this BitMap is corrupt: %s*/\n\n",argv[carg]); carg++; continue;}
   
   if (cint && (sy>0)){
    /*Process intensity map format*/
    t2=0; /*Start position of the image*/
    if (zero>=0) black=zero;
    t4=sy-1;  /*y size*/
    if (t4>16) t4=16;
    printf("/*%s intensity data begins*/\n\n",argv[carg]);

    intmapsc=0;     
    while((t2<sx) && (intmapsc<256)){ /*First process all images for the intensity maps*/
    
     t3=t2+1;
     while(bmp[sxa*(sy-1)+t3]!=black){ /*Seeking to next image's start*/
      t3++;
      if (t3==sx) break;
     }
      
     t3-=t2; /*Width of image*/
      
     if (t3<=16){ /*Suitable: Create intensity map in the temp. workspace*/
      intmapsx[intmapsc]=t3;
      intmapsy[intmapsc]=t4;
      for (t5=0;t5<t4;t5++)
       for (t7=t2;t7<t2+t3;t7++)
        intmaps[(t5<<4)+t7-t2+(intmapsc<<8)]=intmap[bmp[sxa*(sy-t5-2)+t7]];
      intmapsc++; /*Count them!*/
     }
     
     t2+=t3;
    } /*Preprocessing done!*/
    
    /*Now attempt to optimize by looking for identical images & setting copy
    **and mirror flags*/
    for (t1=0;t1<256;t1++) intmapsm[t1]=0; /*Do not copy*/
    t1=1; /*Image pointer*/
    while (t1<intmapsc){ /*Comapre all*/
     t2=0;
     t8=0; /*Matching image found?*/
     while(t2<t1){ /*To all the images before*/
      if (intmapsm[t2]==0){ /*Only if the prev. image is not already a copied one*/
      
       /*If the sizes match, the four mirrorings are possible*/
       if ((intmapsx[t2]==intmapsx[t1]) && (intmapsy[t2]==intmapsy[t1])){
        t7=0xF;
        for (t4=0;t4<intmapsy[t2];t4++)
	 for (t5=0;t5<intmapsx[t2];t5++){
	  if (intmaps[(t1<<8)+(t4<<4)+t5]!=intmaps[(t2<<8)+(t4<<4)                 +t5])                t7&=0xE;
	  if (intmaps[(t1<<8)+(t4<<4)+t5]!=intmaps[(t2<<8)+((intmapsy[t2]-t4-1)<<4)+t5])                t7&=0xD;
          if (intmaps[(t1<<8)+(t4<<4)+t5]!=intmaps[(t2<<8)+((intmapsy[t2]-t4-1)<<4)+intmapsx[t2]-t5-1]) t7&=0xB;
	  if (intmaps[(t1<<8)+(t4<<4)+t5]!=intmaps[(t2<<8)+(t4<<4)                 +intmapsx[t2]-t5-1]) t7&=0x7;
	 }
        if (t7&1){ t8=1; intmapsm[t1]=1;     break; }
        if (t7&2){ t8=1; intmapsm[t1]=1|2;   break; }
        if (t7&4){ t8=1; intmapsm[t1]=1|2|4; break; }
        if (t7&8){ t8=1; intmapsm[t1]=1|  4; break; }
       }
       /*If sizes cross - match, then mirrorxy might be possible*/
       /*Not implemented yet since y size is static across the image, so this
       **has no sense here at all*/
      }
      t2++;
     }
     if (t8) intmapscopy[t1]=t2;
     t1++;
    }
    
    /*Copiings done, now compress intensity map. Here if the image is
    **recangular, it may worth to see if mirror - xy reduces size: because of
    **the linear behavior, this may be possible*/
    
    /*First write no. of images*/
    printf(" %d, /*number of images -1*/\n\n",intmapsc-1);
    osize=1;
    
    for (t1=0;t1<intmapsc;t1++){
    
     if ((intmapsm[t1]&1)==0){ /*If not copied*/

      /*Create intensity map in first temp store*/
      t2=0; /*It's size*/
      for (t5=0;t5<intmapsy[t1];t5++){
       t7=0;
       while (t7<intmapsx[t1]){
        t6=intmaps[(t1<<8)+(t5<<4)+t7]; t8=0;
        while(t6==intmaps[(t1<<8)+(t5<<4)+t7]){ t7++; t8++; if(t7==(intmapsx[t1])) break; }
	intmapst[t2]=(t6<<4)+t8-1;
	t2++;
       }
      }
      
      /*If sizes are equal, mirror to XY, and create a second intensity map
      **in the second temp. store*/
      if (intmapsx[t1]==intmapsy[t1]){
      
       /*Mirror*/
       t4=intmapsx[t1];
       for (t3=0;t3<t4-1;t3++)
        for (t5=t3+1;t5<t4;t5++){
	 t6=intmaps[(t1<<8)+(t3<<4)+t5];
	 intmaps[(t1<<8)+(t3<<4)+t5]=intmaps[(t1<<8)+(t5<<4)+t3];
	 intmaps[(t1<<8)+(t5<<4)+t3]=t6;
	}
	
       /*Create second intensity map, t3: size*/
       t3=0;
       for (t5=0;t5<t4;t5++){
        t7=0;
        while (t7<t4){
         t6=intmaps[(t1<<8)+(t5<<4)+t7]; t8=0;
         while(t6==intmaps[(t1<<8)+(t5<<4)+t7]){ t7++; t8++; if(t7==t4) break; }
 	 intmapst[t3+256]=(t6<<4)+t8-1;
	 t3++;
        }
       }
       
       /*Did it worth?*/
       if (t3<t2){ /*Yes, so set and copy!*/
        intmapsm[t1]|=8;
	for (t4=0;t4<t3;t4++) intmapst[t4]=intmapst[t4+256];
	t2=t3;
       }
       
      } /*OK, done, mirrored if it was necessary*/

     }

     /*Write the two header bytes*/
     osize+=2;
     printf("%s",HexOut(intmapsm[t1]));
     if (intmapsm[t1]&1) t3=intmapscopy[t1];
     else t3=((intmapsx[t1]-1)<<4)+intmapsy[t1]-1;
     printf("%s",HexOut(t3));
     if (intmapsm[t1]&1) printf(" /*mirroring flags & copied image*/\n");
     else printf(" /*head & sizes*/\n");
     
     /*Write the intensity data when necessary*/
     if ((intmapsm[t1]&1)==0){
      osize+=t2;
      t4=0;
      for (t3=0;t3<t2;t3++){
       printf("%s",HexOut(intmapst[t3]));
       t4+=(intmapst[t3]&0xF)+1;
       if (t4==intmapsx[t1]){ printf("\n"); t4=0; }
      }
      printf("/*Intensity map of %d bytes*/\n",t2);
     }
     printf("\n");
     
    }/*All intensity maps written*/

    printf("/*End of %d bytes in total*/\n\n",osize);
    oasize+=osize;
    
    printf("\n");
   }
    
   if (cppc){
   
    osize=2;
    printf("/*%s PPC-PCX data begins*/\n\n",argv[carg]);
    printf("%s",HexOut((sy>>8)& 0xFF));
    printf("%s /*Y size in MSB first format*/\n",HexOut(sy & 0xFF));
   
    /*Position at the image's end, and work backwards like how that crappy
    **BMP format is organized :p */
    bmp+=sxa*(sy-1);
    
    for (t1=0;t1<sy;t1++){
    
     /*Overall behavior of the algorithm:
     **Tries to interpret continous sequences. Reads in input for checking in
     **advance until it turns out that a sequence should be closed, then finds out
     **where the divison between the two sequences should be made.*/
 
     /* t2;     Offset of last checked */
     /* t3;     Offset of >191 byte */
     in=bmp; /* Input line */
     t6=sx;  /* Remaining bytes*/
 
     while(1){ /*Encoding will return when reaching EOD*/
  
      /*Attempt to write simple sequence. Do it until experiencing that:
      **There are 3 similar bytes following each other
      **In the last 255 bytes two >191 bytes appeared
      **The end of sequence is reached*/
      t3=-1;
      t2=0;
      while (1){
       t4=0;
       if (t2==t6) break; /*To exit from loop*/
       for (t5=1;t5<3;t5++){
        if (t2+t5<t6){ if (in[t2]==in[t2+t5]) t4++; }
        else if (((int)in[t2])==transl) t4++;
       }
       if (t4==2) break; /*All were the same, so 3 similar occured!*/
       if (in[t2]>191){
        if ((t3!=-1) && (t3+255>t2)){ t2=t3; break; }
        t3=t2;
       }
       t2++;
      }
      /*Now 't2' holds the length of the sequence to be written out*/
      t6-=t2;
      while (t2>0){
       if ((*in)>191){ printf("%s",HexOut(193)); osize++; } /*One pixel long chunk*/
       printf("%s",HexOut(*in)); osize++;
       t2--; in++;
      }
  
      /*Close the sequence when necessary*/
      if (t6<2){ /*The sortest sequence may be handed over is 2 bytes long,
                 **so anything shorter marks that the data ended. For strict
	         **data encoding, only 2 or longer, or 0 may be t6 here.
	         **(2 can appear when two "big" pixels are at the end)*/
       printf("%s",HexOut(255)); osize++; break;
      }

      /*Test for repeat sequence, start one if yes.
      **It will start if 2 similar pixels follow each other*/
      if ((*in)==in[1]){
  
       t2=0;
       while (1){
        if (t2==t6) break;
        t2++;
        if (t2<t6) if (in[t2-1]!=in[t2]) break; /*Not similar*/
       }
       /*'t2' holds the no. of similar bytes... Write them out!*/
       if (((int)(*in))==transl){ /*Special transparent sequence*/
        if (t2==t6){ /*At the end of the line*/
         printf("%s",HexOut(255)); osize++; break;
        }else{
         while (t2>1){ /*Will trash 1 pixel long to get fed in the others*/
 	  printf("%s",HexOut(252)); osize++; /*Repeat long unused chunk*/
          if (t2>255){
	   printf("%s",HexOut(255)); osize++;
           t6-=255; in+=255; t2-=255;
          }else{
	   printf("%s",HexOut(t2)); osize++;
           t6-=t2; in+=t2; t2=0;
          }
         }
        }
       }else{ /*Not transparent (usual) sequence*/
        while (t2>1){ /*Will trash too short again*/
         if (t2>255){ /*Repeat long*/
	  printf("%s",HexOut(253));
	  printf("%s",HexOut(255));
	  printf("%s",HexOut(*in)); osize+=3;
          t6-=255; in+=255; t2-=255;
         }else if (t2>59){ /*Repeat long (still)*/
	  printf("%s",HexOut(253));
	  printf("%s",HexOut(t2));
	  printf("%s",HexOut(*in)); osize+=3;
	  t6-=t2; in+=t2; t2=0;
         }else{ //Repeat short
	  printf("%s",HexOut(192+t2));
	  printf("%s",HexOut(*in)); osize+=2;
          t6-=t2; in+=t2; t2=0;
         }
        }
       }
   
      }else{
  
       /*Try to write unpacked sequence. Do this until:
       **There are 4 similar bytes following each other treating EOD as similar
       **to transparent.
       **255 byte limit was reached.
       **EOD was reached*/
       t2=0;
       while (1){
        t4=0;
        if (t2==255) break;
        if (t2==t6) break;
        for (t5=1;t5<4;t5++){
         if (t2+t5<t6){ if (in[t2]==in[t2+t5]) t4++; }
         else if (((int)in[t2])==transl) t4++;
        }
        if (t4==3) break; //All were the same, so 4 similar occured!
        t2++;
       }
       /*'t2' holds the number of bytes to write out*/
       if (t2>0){
        printf("%s",HexOut(254));
        printf("%s",HexOut(t2)); osize+=2;
        while(t2>0){
         printf("%s",HexOut(*in)); osize++;
	 in++; t6--; t2--;
        } 
       }

      }
     }

     printf("\n");
     bmp-=sxa;
    }
    printf("\n/*End of %d bytes*/\n\n\n",osize);
    oasize+=osize;
   }

   free(bmporig);
  }
  
  carg++;
 }
 
 if (oasize) printf("\n/*A total of %d bytes were written*/\n\n\n",oasize);
 free (intmaps); /*Free this too when necessary*/
 
 return 0;
 
}
