/*
** levelprc.c
**
** Kye level processor: sets up the objects and map elements according
** to the information read out either from standard Kye levels, or
** RuChKye levels
*/


#include "baseinc.h"
#include "kyemapmn.h"
#include "levelprc.h"


/*Level_KyeBlock
**Translates a standard Kye level block, and returns it as a RuChKye block
**x and y are for x,y coordinates: some objects (auto - rocky, slider) rely
**on it in the original*/
void Level_KyeBlock(byte inchar, adint x, adint y, byte* tchar, byte* rchar, byte* achar)
{
 byte dirs[4]={'u','r','d','l'};
 byte aattr=(((y>>2)&0x3)+'a')-((y>>4)&1)*('a'-'A'); /*For auto - slider / rocky*/
 *rchar='_';                        /*Unused (up) unless necessary to specify*/
 *achar='_';                        /*No attribute unless necessary*/
 switch (inchar){
  case '7': *tchar='w'; *achar='1'; return; /*Wall, round UL*/
  case '8': *tchar='w'; *achar='3'; return; /*Wall, round UL, UR*/
  case '9': *tchar='w'; *achar='2'; return; /*Wall, round UR*/
  case '4': *tchar='w'; *achar='9'; return; /*Wall, round UL, BL*/
  case '5': *tchar='w'; *achar='0'; return; /*Wall*/
  case '6': *tchar='w'; *achar='6'; return; /*Wall, round UR, BR*/
  case '1': *tchar='w'; *achar='8'; return; /*Wall, round BL*/
  case '2': *tchar='w'; *achar='C'; return; /*Wall, round BL, BR*/
  case '3': *tchar='w'; *achar='4'; return; /*Wall, round BR*/
  case 'K': *tchar='k'; return;             /*Kye*/
  case '*': *tchar='*'; *rchar=dirs[((x+10)*(y+30))& 0x3]; return; /*Diamond (rotation is "random")*/
  case 'b': *tchar='b'; *achar='b'; return; /*Box block*/
  case 'B': *tchar='b'; *achar='r'; return; /*Round block*/
  case 'e': *tchar='e'; *achar='b'; return; /*Eatable (box) block*/
  case 'u': *tchar='s'; *rchar='u'; return; /*Slider U*/
  case 'r': *tchar='s'; *rchar='r'; return; /*Slider R*/
  case 'd': *tchar='s'; *rchar='d'; return; /*Slider D*/
  case 'l': *tchar='s'; *rchar='l'; return; /*Slider L*/
  case 'U': *tchar='m'; *rchar='u'; return; /*Mover U*/
  case 'R': *tchar='m'; *rchar='r'; return; /*Mover R*/
  case 'D': *tchar='m'; *rchar='d'; return; /*Mover D*/
  case 'L': *tchar='m'; *rchar='l'; return; /*Mover L*/
  case '^': *tchar='r'; *rchar='u'; return; /*Rocky U*/
  case '>': *tchar='r'; *rchar='r'; return; /*Rocky R*/
  case 'v': *tchar='r'; *rchar='d'; return; /*Rocky D*/
  case '<': *tchar='r'; *rchar='l'; return; /*Rocky L*/
  case 'c': *tchar='c'; *achar='c'; return; /*Clocker*/
  case 'a': *tchar='c'; *achar='a'; return; /*Anti - clocker*/
  case 'A': *tchar='S'; *rchar=dirs[(x & 1)<<1]; *achar=aattr; return; /*Auto - slider*/
  case 'F': *tchar='R'; *rchar=dirs[(x & 1)<<1]; *achar=aattr; return; /*Auto - rocky*/
  case 's': *tchar='p'; *rchar='u'; return; /*Sticky (puller) up & down*/
  case 'S': *tchar='p'; *rchar='r'; return; /*Sticky left & right*/
  case 'T': *tchar='v'; *achar='t'; *rchar=dirs[((x+11)*(y+35))& 0x3]; return; /*Virus: Twister*/
  case 'C': *tchar='v'; *achar='b'; *rchar=dirs[((x+12)*(y+34))& 0x3]; return; /*Virus: Blob*/
  case 'E': *tchar='v'; *achar='g'; *rchar=dirs[((x+13)*(y+33))& 0x3]; return; /*Virus: Gnasher*/
  case '~': *tchar='v'; *achar='w'; *rchar=dirs[((x+14)*(y+32))& 0x3]; return; /*Virus: Worm*/
  case '[': *tchar='v'; *achar='s'; *rchar=dirs[((x+15)*(y+31))& 0x3]; return; /*Virus: Spike*/
  case 'H': *tchar='h'; return;             /*Black hole*/
  case 'i': *tchar='d'; *rchar='u'; *achar='k'; return; /*Door (invisible direction) for Kye to up*/
  case 'f': *tchar='d'; *rchar='r'; *achar='k'; return; /*Door for Kye to right*/
  case 'h': *tchar='d'; *rchar='d'; *achar='k'; return; /*Door for Kye to down*/
  case 'g': *tchar='d'; *rchar='l'; *achar='k'; return; /*Door for Kye to left*/
  case '}': *tchar='t'; *achar='3'; return; /*Timer 3*/
  case '|': *tchar='t'; *achar='4'; return; /*Timer 4*/
  case '{': *tchar='t'; *achar='5'; return; /*Timer 5*/
  case 'z': *tchar='t'; *achar='6'; return; /*Timer 6*/  
  case 'y': *tchar='t'; *achar='7'; return; /*Timer 7*/
  case 'x': *tchar='t'; *achar='8'; return; /*Timer 8*/
  case 'w': *tchar='t'; *achar='9'; return; /*Timer 9*/
  default:  *tchar=' '; return;             /*Empty block*/
 }
}


/*Level_ParseRot
**Parses rotation character and returns 0-3 appropriately*/
adint Level_ParseRot(byte rchar)
{
 switch (rchar){
  case 'L': case 'l':
  case 'E': case 'e':
  case '6': case '>': return 3;
  case 'D': case 'd': case 'B': case 'b':
  case 'S': case 's':
  case '2': case 'v': case 'V': return 2;
  case 'R': case 'r':
  case 'W': case 'w':
  case '4': case '<': return 1;
  default: return 0;  
 }
}

/*Level_SkipLine
**Skips a line in the input (reads until passing over \n)*/
void Level_SkipLine(byte (*feeder) ())
{
 byte t=feeder();
 while ((t!='\n') && (t!=0xFF)) t=feeder();
}

/*Level_GetNumLine
**Parses a line to it's end which supposed to be numeric, and returns the
**number could be extracted, 0 if nothing numeric, 0xFFFF if the RuChKye Level
**string was experienced. Feeder is the function which feeds it with
**characters, 0xFF on eof*/
adint Level_GetNumLine(byte (*feeder) ())
{
 byte ruchkye[13]={'r','u','c','h','k','y','e',' ','l','e','v','e','l'};
 byte nextchar;
 adint onum=0;
 adint charc=0;

 while(1){
  nextchar=feeder();
  if ((nextchar>=58) && (charc==0)) /*Not numeric, maybe ruchkye*/
   while(1){
    if ((nextchar|0x20)!=ruchkye[charc]){
     if (nextchar!='\n') Level_SkipLine(feeder);
     return 0;
    }
    if(charc==12){
     if (nextchar!='\n') Level_SkipLine(feeder);
     return 0xFFFF; /* "RuChKye Level" string got!*/
    }
    charc++;
    nextchar=feeder();
   }
  if ((nextchar>=58) || (nextchar<48)){
   if (nextchar!='\n') Level_SkipLine(feeder);
   return onum;
  }else{
   onum*=10;
   onum+=(nextchar-48);
   charc++;
  }
 }
}


/*Level_RuChKyeBlock
**Translates a RuChKye level block
**tchar: Type selector
**rchar: Rotation
**  U/u/T/t/N/n/8/^/A: Up (Top, North)
**  L/l/    E/e/6/>:   Left (East)
**  D/d/B/b/S/s/2/v/V: Down (Bottom, South)
**  R/r/    W/w/4/<:   Right (West)
**achar: Attributes
**xi and yi are the coordinates in the input, mx and my are where the block
**supposed to go. These are seperated to properly interpret original kye
**style type specifications (regardless to where the map will go on the
**playing field)
**returns 0 unless Kye was placed (1), or a diamond (2)
*/
adint Level_RuChKyeBlock(byte tchar, byte rchar, byte achar, adint xi, adint yi, adint mx, adint my)
{
 adint rot;
 adint t;

 Level_RuChKyeBlock_Restart:
 /*To restart parsing when some combination has to be converted*/
 
 rot=Level_ParseRot(rchar);
 
 switch (tchar){
  case '-': case '_': /*Original Kye*/
   Level_KyeBlock(achar,xi,yi,&tchar,&rchar,&achar); /*Parse original Kye block*/
   goto Level_RuChKyeBlock_Restart; /*Restart with the new attributes*/
  case 'w': case 'W': /*Wall*/
   t=0;
   if ((achar>='0') && (achar<='9')) t=achar-'0';
   else if ((achar>='A') && (achar<='F')) t=achar-'A'+10;
   else if ((achar>='a') && (achar<='f')) t=achar-'a'+10;
   Kye_BlockSet(mx,my,0x10 | t);
   return 0;
  case 'd': case 'D':
   t=rot;
   switch (achar){
    case 'A': t|=0x10; /*Visible flag*/
    case 'a': break;
    case 'K': t|=0x10;
    default:  t|=0x04; /*Kye only flag*/
   }
   Kye_BlockSet(mx,my,0x20 | t);
   return 0;
  case 'k': case 'K': /*Kye*/
   Kye_NewObject(mx,my,KYE_O_KYE,0,0);
   return 1;
  case '*': /*Diamond*/
   Kye_NewObject(mx,my,KYE_O_DIAMOND,rot,0);
   return 2;
  case 'B': case 'b': /*Pushable block*/
   if ((achar=='r') || (achar=='R')) Kye_NewObject(mx,my,KYE_O_RBLOCK,0,0);
   else Kye_NewObject(mx,my,KYE_O_BLOCK,0,0);
   return 0;
  case 'E': case 'e': /*Eatable block*/
   if ((achar=='r') || (achar=='R')) Kye_NewObject(mx,my,KYE_O_ERBLOCK,0,0);
   else Kye_NewObject(mx,my,KYE_O_EBLOCK,0,0);
   return 0;
  case 's': /*Slider*/
   Kye_NewObject(mx,my,KYE_O_SLIDER,rot,0);
   return 0;
  case 'r': /*Rocky*/
   Kye_NewObject(mx,my,KYE_O_ROCKY,rot,0);
   return 0;
  case 'm': case 'M': /*Mover*/
   Kye_NewObject(mx,my,KYE_O_MOVER,rot,0);
   return 0;
  case 'S': case 'R': /*Auto - slider and rocky*/
   switch (achar){
    case 'a': t=3; break;
    case 'b': t=2; break;
    case 'c': t=1; break;
    case 'd': t=0; break;
    case 'A': t=7; break;
    case 'B': t=6; break;
    case 'C': t=5; break;
    case 'D': t=4; break;
    default: /*Make slider / rocky of them*/
     tchar|=0x20; /*Lowercase*/
     goto Level_RuChKyeBlock_Restart; /*Restart to make those objects instead*/
   }
   if (tchar=='S') Kye_NewObject(mx,my,KYE_O_ASLIDER,rot,t);
   else Kye_NewObject(mx,my,KYE_O_AROCKY,rot,t);
   return 0;
  case 'c': case 'C':
   if ((achar>='0') && (achar<='9')){
    tchar='t'; /*Timer block*/
    goto Level_RuChKyeBlock_Restart; /*Restart again*/
   }
   if ((achar=='c') || (achar=='C')) Kye_NewObject(mx,my,KYE_O_CLOCKER,rot | 4,0);
   else Kye_NewObject(mx,my,KYE_O_CLOCKER,rot,0);
   return 0;
  case 't': case 'T':
   if ((achar<'0') || (achar>'9')) achar='3';
   Kye_NewObject(mx,my,KYE_O_TBLOCK,0,((((((adint)achar)-'0'+1)*25)+5)<<8) | KYE_NEW_TROT);
   return 0;
  case 'p': case 'P':
   Kye_NewObject(mx,my,KYE_O_MAGNET,rot,0);
   return 0;
  case 'h': case 'H':
   Kye_NewObject(mx,my,KYE_O_BLACK,0,0);
   return 0;
  case 'v': case 'V':
   switch (achar){
    /* 'B': t=0; Default!*/
    case 'T': case 't': t=3; break;
    case 'S': case 's': t=4; break;
    case 'G': case 'g': t=2; break;
    case 'W': case 'w': case '~': t=1; break;
    default: t=0;
   }
   Kye_NewObject(mx,my,KYE_O_VIRUS,rot | (t<<2),0);
   return 0;
  default: return 0; /*Empty or not yet implemented*/
 }
}


/*Outputs an original Kye level on the middle of the playing field.
**The feeder should start feeding at the first line of the level data
**returns 0 if the level was all OK, 1 if it is buggy, but might be playable,
**2 if it is a complete mess (no or more Kyes, or no diamonds)*/
/*(Original Kye had 30x20 maps)*/
adint Level_Kye(byte (*feeder) ())
{
 adint diam=0;
 adint kye=0;
 adint bug=0;
 adint t;
 adint x=0,y=0,mx=(KYE_MX-30)>>1,my=(KYE_MY-20)>>1;
 byte nextch=feeder();
 
 /*First jump through the first few empty lines if any (for example if the
 **feeder was left before a new line)*/
 while(nextch<32) nextch=feeder();
 
 while(1){
  if (nextch==0xFF) break;
  if (nextch=='\n'){ /*Next line*/
   if (x!=30) bug=1;
   x=0; mx=(KYE_MX-30)>>1; y++; my++;
   if (y>=20) break;
  }else if (nextch<32){ /*Ignore control characters*/
  }else if (x<30){
   t=Level_RuChKyeBlock('_','_',nextch,x,y,mx,my); /*Like if it was a Kye block in a RuChKye map*/
   if (t==1) kye++;
   if (t==2) diam++;
   x++; mx++;
  }else bug=1;
  nextch=feeder();
 }
 
 if ((kye!=1) || (diam==0)) return 2; /*Unplayable!*/
 if ((bug) || (y<20)) return 1; /*Buggy, but maybe can go*/
 return 0; /*OK*/
}

/*Similar to the above for RuChKye level data
**sx and sy are the map sizes. However in RuChKye maps it is allowed for a line
**to be shorter than the map's size*/
adint Level_RuChKye(byte (*feeder) (), adint sx, adint sy)
{
 adint diam=0;
 adint kye=0;
 adint bug=0;
 adint t;
 adint x=0,y=0,mx=(KYE_MX-sx)>>1,my=(KYE_MY-sy)>>1;
 byte nextch=feeder();
 
 if ((sx>KYE_MX) || (sy>KYE_MY)) return 2; /*Too big map!*/
 
 /*First jump through the first few empty lines if any (for example if the
 **feeder was left before a new line)*/
 while(nextch<32) nextch=feeder();
 
 while(1){
  if (nextch==0xFF) break;
  if (nextch=='\n'){ /*Next line*/
   if (x>sx) bug=1;
   x=0; mx=(KYE_MX-sx)>>1; y++; my++;
   if (y>=sy) break;
  }else if (nextch<32){ /*Ignore control characters*/
  }else if (x<sx){
   t=Level_RuChKyeBlock(nextch,feeder(),feeder(),x,y,mx,my);
   if (t==1) kye++;
   if (t==2) diam++;
   x++; mx++;
  }else bug=1;
  nextch=feeder();
 }
 
 if ((kye!=1) || (diam==0)) return 2; /*Unplayable!*/
 if ((bug) || (y<sy)) return 1; /*Buggy, but maybe can go*/
 return 0; /*OK*/
}
